(in-package :ucw)

(defclass range-cursor ()
  ((data :accessor range-cursor.data
                 :initarg :data
                 :documentation
                 "Data to range.")
   (chunk-size :accessor range-cursor.chunk-size
               :initarg :chunk-size
               :type integer
               :initform 20
               :documentation
               "Maximum number of elements in one chunk.")
   (offset :accessor range-cursor.offset
           :type integer
           :initarg :offset
           :initform 0
           :documentation
           "Current offset in chunks.")
   (recent-fetch-size :reader range-cursor.recent-fetch-size
                     :initform nil
                     :documentation
                     "Number of recently retrieved data elements."))
  (:documentation "Read chunks of elements from data at arbitrary offset."))

(defgeneric range-cursor.data-size (range-cursor data)
  (:method ((rc range-cursor) (rd sequence))
    (length rd))
  (:documentation "Return total number of elements in the DATA."))

(defgeneric range-cursor.fetch-data-chunk (range-cursor data offset length)
  (:method ((rc range-cursor) (data sequence) (offset integer) (length integer))
    (subseq data offset (+ offset length)))
  (:documentation "Read LENGTH elements from DATA starting from
OFFSET(excluding OFFSET element).
Return a list of elements."))

(defgeneric range-cursor.delete-data-item (range-cursor data item index)
  (:method ((rc range-cursor) (data sequence) item (index integer))
    (declare (ignore item))
    (when (<= index (length data))
      (setf (range-cursor.data rc)
            (delete-if (constantly t) data :start (1- index) :count 1))
      t))
  (:documentation "Delete ITEM with INDEX from cursor related DATA."))

(defgeneric range-cursor.add-data-item (range-cursor data item)
  (:method ((rc range-cursor) (data sequence) item)
    (push item (range-cursor.data rc))
    t)
  (:documentation "Add ITEM with INDEX to cursor related DATA."))

(defgeneric range-cursor.range-size (range-cursor)
  (:documentation "Return the size of the set in chunks."))

(defmethod range-cursor.range-size ((rc range-cursor))
  (with-slots (data chunk-size) rc
    (multiple-value-bind (q rem)
        (floor (range-cursor.data-size rc data) chunk-size)
      (if (zerop rem) q (1+ q)))))

(defmethod range-cursor.have-previous-p ((rc range-cursor))
  (> (range-cursor.offset rc) 0))

(defmethod range-cursor.have-next-p ((rc range-cursor))
  (aif (range-cursor.recent-fetch-size rc)
       (> it (range-cursor.chunk-size rc))
       t ; no reads yet
       ))

(defgeneric range-cursor.fetch-chunk (range-cursor op)
  (:documentation "Retrieve chunk from the data.
Returns a list of elements."))

(defmethod range-cursor.fetch-chunk :before ((rc range-cursor) op)
  (with-slots (offset recent-fetch-size) rc
    (setf offset
          (ecase op
            (:first 0)
            (:last (max (1- (range-cursor.range-size rc)) 0))
            (:next (if (and recent-fetch-size (range-cursor.have-next-p rc))
                       (1+ offset)
                       offset))
            (:previous (if (range-cursor.have-previous-p rc)
                           (1- offset)
                           offset))
            (:current (max 0 offset))))))

(defmethod range-cursor.fetch-chunk ((rc range-cursor) op)
  (declare (ignore op))
  (with-slots (chunk-size recent-fetch-size offset data) rc
    (let ((items (range-cursor.fetch-data-chunk rc data (* offset chunk-size) (1+ chunk-size))))
      (setf recent-fetch-size (length items))
      (if (> recent-fetch-size chunk-size)
          (butlast items)
          items))))

(defmethod range-cursor.delete-item ((rc range-cursor) item (index integer))
  "Delete ITEM with INDEX from cursor related DATA."
  (range-cursor.delete-data-item rc (range-cursor.data rc) item index))

(defmethod range-cursor.add-item ((rc range-cursor) item)
  "Add ITEM to cursor related DATA."
  (range-cursor.add-data-item rc (range-cursor.data rc) item))

(define-ie-type (range-set range-set-element) ((ie-type set))
  ((offset :accessor range-set.offset
           :initform 0
           :type integer
           :backtrack t
           :documentation
           "Which of the windows we're currently looking at.")
   (window-size :accessor range-set.window-size
                :initform 20
                :backtrack t
                :initarg :window-size)
   (cursor :accessor range-set.cursor
           :type range-cursor))
  (:metaclass standard-component-class)
  (:documentation "Component for showing the user a set of data one \"window\" at a time.

The data set is presented one \"window\" at a time with links to
the the first, previous, next and last window. Each window shows
at most WINDOW-SIZE elements of the data.
Cursor is created as instance of CURSOR-CLASS initarg with
DATA initarg as additional keyword argument to make-instance"))

(defmethod shared-initialize :after ((rs range-set-element) slot-names
                                     &key (cursor-class 'range-cursor) data
                                     &allow-other-keys)
  (declare (ignore slot-names))
  (range-set.init-cursor rs cursor-class data)
  (with-slots (cursor offset window-size item-viewer) rs
    (setf (range-cursor.offset cursor) offset
          (range-cursor.chunk-size cursor) window-size)
    (dolist (sn (slot-elements item-viewer))
      (setf (parent (slot-value item-viewer sn)) rs))))

(defmethod range-set.init-cursor ((rs range-set-element) cursor-class data)
  (setf (range-set.cursor rs) (make-instance cursor-class :data data)))

(defmethod (setf range-set.offset) :after (new (rs range-set-element))
  (setf (range-cursor.offset (range-set.cursor rs)) new))

(defmethod (setf range-set.window-size) :after (new (rs range-set-element))
  (setf (range-cursor.chunk-size (range-set.cursor rs)) new))

(defmethod range-set.have-previous-p ((rs range-set-element))
  "Returns true if we have a window before the current one."
  (range-cursor.have-previous-p (range-set.cursor rs)))

(defmethod range-set.have-next-p ((rs range-set-element))
  "Returns true if we have a window after the current one."
  (range-cursor.have-next-p (range-set.cursor rs)))

(defmethod range-set.fetch-items ((rs range-set-element) op)
  (prog1
      (range-cursor.fetch-chunk (range-set.cursor rs) op)
    (setf (slot-value rs 'offset)
          (range-cursor.offset (range-set.cursor rs)))))

(defmethod scroll ((rs range-set-element) op)
  (setf (items rs) 
        (do ((items (range-set.fetch-items rs op)
                    (range-set.fetch-items rs :previous)))
            ((or items (zerop (range-set.offset rs))) items))))

(defaction scroll-start ((rs range-set-element))
  (scroll rs :first))

(defaction scroll-end ((rs range-set-element))
  (scroll rs :last))
  
(defaction scroll-forward ((rs range-set-element) &optional (n 1))
  (cond ((= 1 n) (scroll rs :next))
        ((zerop n) (scroll rs :current))
        (t (scroll-to-page rs (+ (range-set.offset rs) n)))))

(defaction scroll-backward ((rs range-set-element) &optional (n 1))
  (cond ((= 1 n) (scroll rs :previous))
        ((zerop n) (scroll rs :current))
        (t (scroll-to-page rs (- (range-set.offset rs) n)))))

(defaction scroll-to-page ((rs range-set-element) window-number)
  (setf (range-set.offset rs) window-number)
  (scroll rs :current))

(defaction delete-item ((rs range-set-element) index-in-window)
  (if (<= index-in-window (length (items rs)))
     (with-slots (offset cursor window-size) rs
       (when (range-cursor.delete-item cursor
                                       (nth (1- index-in-window) (items rs))
                                       (+ (* offset window-size)
                                          index-in-window))
         (scroll rs :current)
         t))))

(defaction add-item ((rs range-set-element) item)
  (when (range-cursor.add-item (range-set.cursor rs) item)
    (with-slots (items window-size)
        rs
      (setf items (cons item
                        (if (< (length items) window-size)
                          items
                          (butlast items)))))))

(defmethod render-value-box :after ((view ieview) (rs range-set-element) value)
  (declare (ignore value))
  (let* ((active-style "padding-left:10px; padding-right:10px;")
        (inactive-style (strcat active-style " color:black;")))
  (<:div :class "ucw-composite"
         :style " text-align: center; padding-bottom:10px;"
    (if (range-set.have-previous-p rs)
        (progn
          (<ucw:a :style active-style :action (scroll-start rs)
                  (<:as-html "<<"))
          (<ucw:a :style active-style :action (scroll-backward rs)
                  (<:as-html "<")))
        (progn
          (<:a :style inactive-style "<<")(<:a :style inactive-style "<")))
    (if (range-set.have-next-p rs)
        (progn
          (<ucw:a :style active-style :action (scroll-forward rs)
                  (<:as-html ">"))
          (<ucw:a :style active-style :action (scroll-end rs)
                  (<:as-html ">>")))
        (progn
          (<:a :style inactive-style ">")(<:a :style inactive-style ">>"))))))