;;;; -*- lisp -*-

(in-package :it.bese.ucw)

;;;; ** STANDARD-REQUEST-CONTEXT

(defmethod context.window-component ((context request-context))
  (frame.window-component (context.current-frame context)))

(defmethod (setf context.window-component) ((component component) (context request-context))
  (setf (frame.window-component (context.current-frame context))
        component))

(defmethod call-callbacks ((context standard-request-context))
  "Execute all the callback lambda in CONTEXT's request.

Simply goes through the request's params and, for every param
which isn't +action-parameter-name+, +frame-parameter-name+ or
+session-parameter-name+, looks up and call the associated lambda
in the current frame."
  (map-parameters (context.request context)
                  (lambda (param value)
                    (unless (member param (list +action-parameter-name+
                                                +frame-parameter-name+
                                                +session-parameter-name+)
                                    :test #'string=)
                      (call-callback (context.current-frame context) param value)))))

(defmethod action-request-p ((context standard-request-context))
  "Returns true if the request represents an action (as opposed to entry-point).

The criteria:

action-request iff (context's request has a frame-id and an action-id)
                   and (the frame-id and the action-id exist)
"
  (find-action-id context))

(defmethod find-session-id ((context standard-request-context))
  "Returns the value of the GET or POST parameter named +session-parameter-name+"
  (get-parameter (context.request context) +session-parameter-name+))

(defmethod find-frame-id ((context standard-request-context))
  (get-parameter (context.request context) +frame-parameter-name+))

(defmethod find-action-id ((context standard-request-context))
  (get-parameter (context.request context) +action-parameter-name+))

(defmethod clear-context ((context standard-request-context))
  "Undos (as much as possible) everything that's happened to CONTEXT."
  (clear-response (context.response context))
  context)

;;;; *** DUMMY-REQUEST-CONTEXT

;;;; This context serves as a testing/debugging aid. It allows you to
;;;; create and call components from the repl.

(defclass dummy-request-context (standard-request-context)
  ())

(defclass dummy-request (httpd-request)
  ((query-path :initform "/DUMMY/index.ucw" :initarg :query-path)))

(defclass dummy-response (httpd-response)
  ()
  (:default-initargs :network-stream *trace-output*))

(defun make-dummy-context ()
  (let* ((dummy-app (or *default-application*
                        (make-instance 'standard-application
                                       :url-prefix "/DUMMY/")))
         (*context* (make-request-context dummy-app
                                          (make-instance 'dummy-request)
                                          (make-instance 'dummy-response)))
         (session (make-new-session dummy-app))
         (frame (make-new-frame session)))
    (setf (session.current-frame session) frame
          (context.session *context*) session)
    *context*))

(defcomponent dummy-root-component (window-component)
  ())

(defmethod render ((comp dummy-root-component))
  (<:as-html "DUMMY COMPONENT"))

(defmacro with-dummy-context ((&key (render t)
                                    (action t))
                              &body body)
  `(let* ((*context* (make-dummy-context))
          (self (make-instance 'dummy-root-component
                               :place (make-place
                                       (context.window-component *context*)))))
     (setf (context.window-component *context*) self)
     (multiple-value-prog1
         ,(if action
              `(with-call/cc ,@body)
              `(progn ,@body))
       (when ,render
         (render (context.window-component *context*))
         (shutdown (context.response *context*))))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Copyright (c) 2003-2005 Edward Marco Baringer
;;; All rights reserved. 
;;; 
;;; Redistribution and use in source and binary forms, with or without
;;; modification, are permitted provided that the following conditions are
;;; met:
;;; 
;;;  - Redistributions of source code must retain the above copyright
;;;    notice, this list of conditions and the following disclaimer.
;;; 
;;;  - Redistributions in binary form must reproduce the above copyright
;;;    notice, this list of conditions and the following disclaimer in the
;;;    documentation and/or other materials provided with the distribution.
;;; 
;;;  - Neither the name of Edward Marco Baringer, nor BESE, nor the names
;;;    of its contributors may be used to endorse or promote products
;;;    derived from this software without specific prior written permission.
;;; 
;;; THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
;;; "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
;;; LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
;;; A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT
;;; OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
;;; SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
;;; LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
;;; DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
;;; THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
;;; (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
;;; OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
